package engine

import (
	"context"
	"crypto/rand"
	"encoding/hex"
	"errors"
	"fmt"
	"io"
	"time"

	pb "git.autistici.org/ai3/tools/enq/proto"
	"google.golang.org/protobuf/types/known/emptypb"
	"google.golang.org/protobuf/types/known/timestamppb"
)

// SubmitTask creates an individual task on this shard.
func (s *Engine) SubmitTask(ctx context.Context, spec *pb.TaskSpec) (*emptypb.Empty, error) {
	err := s.db.WithTx(ctx, func(tx databaseTx) error {
		if err := tx.AddTask(spec); err != nil {
			return err
		}
		return nil
	})
	return emptyResponse, err
}

// Submit a task tree to the global queue service. The list of tasks
// must represent a single, connected, tree. This function acts as a
// proxy, and will call other shards to submit individual tasks.
//
// In order to express task relationships, IDs in the submitted task
// tree are just placeholders, where only the *shard* part is
// significant: the unique ID will be randomly generated by this
// function and replaced.
//
// Since we don't have global transactions, the order of individual
// task creation is important: this function will proceed from root to
// the leaves, so as to make sure that downstream tasks always exists
// (otherwise we'd risk errors when propagating results, should a task
// be run before we complete the creation of the rest of the tree).
//
func (s *Engine) Submit(ctx context.Context, req *pb.SubmitRequest) (*pb.SubmitResponse, error) {
	if err := validateTasks(req.Tasks); err != nil {
		return nil, err
	}
	specs, err := rewriteTaskSpecs(req.Tasks)
	if err != nil {
		return nil, err
	}
	if len(specs) == 0 {
		return nil, errors.New("no tasks specified")
	}

	rootID := specs[len(specs)-1].TaskId

	err = s.db.WithTx(ctx, func(tx databaseTx) error {
		for _, spec := range specs {
			peer, err := s.net.GetPeer(ctx, spec.TaskId.Shard)
			if err != nil {
				return err
			}
			_, err = peer.SubmitTask(ctx, spec)
			if err != nil {
				return err
			}
		}
		return nil
	})
	return &pb.SubmitResponse{
		RootTaskId: rootID,
	}, err
}

// Validate task specs and fill in defaults.
func validateTasks(tasks []*pb.TaskSpec) error {
	for _, task := range tasks {
		if err := task.Valid(); err != nil {
			return fmt.Errorf("invalid task %s: %w", task.TaskId.ToString(), err)
		}
		task.SetDefaults()
	}
	return nil
}

// Rewrite and reorder the list of tasks specs so that:
//
// - IDs are replaced with unique ones (maintaining the shard)
//
// - tasks are ordered depth-first, so that no task references a task
//   that has not been created yet (useful in case of partial failure)
//
// Also runs sanity checks on the tree structure to ensure it conforms
// to our expectations.
func rewriteTaskSpecs(specs []*pb.TaskSpec) ([]*pb.TaskSpec, error) {
	// Assign new IDs to each task, build a map of old ids -> new
	// ids, and a map of all the reverse dependencies that will be
	// used later to find the root task.
	tmpIDMap := make(map[string]*pb.TaskSpec)
	realIDs := make(map[string]*pb.ShardedID)
	deps := make(map[string]map[string]struct{})
	for _, spec := range specs {
		tmpIDMap[spec.TaskId.ToString()] = spec
		realIDs[spec.TaskId.ToString()] = newUniqueID(spec.TaskId.Shard)
		for _, arg := range spec.GetTaskArgs() {
			m, ok := deps[arg.TaskId.ToString()]
			if !ok {
				m = make(map[string]struct{})
				deps[arg.TaskId.ToString()] = m
			}
			// if _, ok := m[spec.TaskId.ToString()]; ok {
			// 	return nil, fmt.Errorf("task %s has more than one downstream dependency", spec.TaskId.ToString())
			// }
			m[spec.TaskId.ToString()] = struct{}{}
		}
	}

	// Find the root task (no reverse deps), verify that there is only one.
	var rootID *pb.ShardedID
	for _, spec := range specs {
		_, ok := deps[spec.TaskId.ToString()]
		if !ok {
			if rootID == nil {
				rootID = spec.TaskId
			} else {
				return nil, errors.New("more than one root task in the tree")
			}
		}
	}

	// Navigate the tree depth-first to produce the final
	// ordering. Rewrite all task IDs with the real ones (both in
	// the TaskSpec structure itself and in taskref arguments) at
	// this stage.
	var out []*pb.TaskSpec
	var df func(id *pb.ShardedID) error
	df = func(id *pb.ShardedID) error {
		spec, ok := tmpIDMap[id.ToString()]
		if !ok {
			return fmt.Errorf("referenced task %s not found", id.ToString())
		}
		spec.TaskId = realIDs[id.ToString()]
		for i := 0; i < len(spec.Args); i++ {
			if spec.Args[i].Type == pb.ArgumentSpec_TASKREF {
				if err := df(spec.Args[i].TaskId); err != nil {
					return err
				}
				// Set the DownstreamTaskID of the referenced
				// task to the current task ID.
				tmpIDMap[spec.Args[i].TaskId.ToString()].DownstreamTaskId = spec.TaskId
				spec.Args[i].TaskId = realIDs[spec.Args[i].TaskId.ToString()]
			}
		}
		out = append(out, spec)
		return nil
	}
	if err := df(rootID); err != nil {
		return nil, err
	}
	if len(out) != len(specs) {
		return nil, errors.New("unused tasks in the tree")
	}

	return out, nil
}

// Query the database for information about a specific task on this shard.
func (s *Engine) Query(ctx context.Context, req *pb.QueryRequest) (*pb.TaskInfo, error) {
	if s.shard == req.TaskId.Shard {
		var info pb.TaskInfo
		err := s.db.WithTx(ctx, func(tx databaseTx) error {
			// Aggregate all the TaskInfo components.
			spec, err := tx.GetTaskSpec(req.TaskId)
			if err != nil {
				return err
			}
			info.Spec = spec

			state, err := tx.GetTaskState(req.TaskId)
			if err != nil {
				return err
			}
			info.State = state

			if t, err := tx.GetTaskCreatedAt(req.TaskId); err == nil {
				info.CreatedAt = timestamppb.New(t)
			}

			return nil
		})
		return &info, err
	}

	// Forward remote call.
	peer, err := s.net.GetQueue(ctx, req.TaskId.Shard)
	if err != nil {
		return nil, err
	}
	return peer.Query(ctx, req)
}

// GetLogs streams logs for a task (logs can grow large).
func (s *Engine) GetLogs(req *pb.GetLogsRequest, stream pb.Queue_GetLogsServer) error {
	// Set a max running time of 2 minutes.
	ctx, cancel := context.WithTimeout(context.Background(), 2*time.Minute)
	defer cancel()

	if s.shard == req.TaskId.Shard {
		return s.db.WithTx(ctx, func(tx databaseTx) error {
			r, err := tx.GetLogs(req.TaskId)
			if err != nil {
				return err
			}
			defer r.Close()
			buf := make([]byte, 64*1024)
			for {
				n, err := r.Read(buf)
				if err == io.EOF {
					return nil
				}
				if err != nil {
					return err
				}
				if err := stream.Send(&pb.LogChunk{Data: buf[:n]}); err != nil {
					return err
				}
			}
		})
	}

	// Forward packets from the remote stream to the client.
	peer, err := s.net.GetQueue(ctx, req.TaskId.Shard)
	if err != nil {
		return err
	}
	upstream, err := peer.GetLogs(ctx, req)
	if err != nil {
		return err
	}
	for {
		chunk, err := upstream.Recv()
		if err == io.EOF {
			return nil
		}
		if err != nil {
			return err
		}
		if err := stream.Send(chunk); err != nil {
			return err
		}
	}
}

func newUniqueID(shard string) *pb.ShardedID {
	var b [8]byte
	if _, err := io.ReadFull(rand.Reader, b[:]); err != nil {
		panic(err)
	}
	return &pb.ShardedID{
		Shard: shard,
		Id:    hex.EncodeToString(b[:]),
	}
}
